/***************************************************************************
 *
 * Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: lineio.c,v 1.2 2000/08/01 10:10:36 bazsi Exp $
 *
 ***************************************************************************/

#include <zorp/lineio.h>

#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <glib.h>

#include <openssl/ssl.h>


#define BUFFER_LEN 256


/*
   Beolvas az atadott SSL kapcsolatbol egy sort, es a szokozoknel szetvagja
   tokenekre, amiket felfuz egy ujonnan allokalt GListbe. A visszateresi
   ertek ez a lista. A hivo fv felelossege a GList felszabaditasa.
   Hiba eseten a visszateresi ertek NULL.
   
   Max sorhossz BUFFER_LEN byte;
*/

GList *read_token_line(SSL *ssl)
{
	GList *list=NULL;
	char buffer[BUFFER_LEN];
	char *element,a;
	int ret,cnt,vedett;
	

	cnt=0;
	while(cnt<BUFFER_LEN){
		ret=SSL_read(ssl,&a,1);
		if(ret<0){
			g_list_free(list);
			return NULL;
		}
		if(ret!=1) a='\n';
		if((a=='\n' || a==' ') && !vedett){
			buffer[cnt]=0;
			element=g_new0(char,cnt+1);
			strncpy(element,buffer,cnt);
			list=g_list_append(list,element);
			cnt=0;
		}else if(a=='\\' && !vedett){
			vedett=1;
		}else{
			buffer[cnt++]=a;
			vedett=0;
		}
		if(a=='\n')
			break;
		
	}
	return list;
}

/*
   2 parameteres free(). A masodik parameter nem szamit semmit,
   csak igy ahogy van kell az egesz a kovetkezo fv-nek.
*/
static void b_free(gpointer str, gpointer a)
{
	g_free(str);
}

/*
   felszabaditja a fenti fuggveny altal visszaadott Lista osszes elemet,
   es magat a listat.
*/
void free_g_list_of_strings(GList *list)
{
	g_list_foreach(list,b_free,NULL);
	g_list_free(list);
}

/*
   Beolvas egy sort. Ujonnan allokal helyet neki, es azt adja vissza.
   Az allokalt hely felszabaditasa a hivo fv felelossege.
   
   Max sorhossz BUFFER_LEN byte;
*/
char *read_line(SSL *ssl)
{
	char buffer[BUFFER_LEN],a;
	char *line;
	int ret,cnt;
	
	cnt=0;
	while(1){
		ret=SSL_read(ssl,&a,1);
		if(ret<0)
			return NULL;
		if(a=='\n'){
			buffer[cnt]=0;
			line=g_new0(char,cnt+1);
			strncpy(line,buffer,cnt);
			break;
		}
		buffer[cnt++]=a;
	}
	return line;
}


/*
   beolvas egy parancsot, ami egy token lista. Visszater a beolvasott
   tokenek listajaval, ha az elso token szerepel a cmds tombben, es
   megfelelo szamu argumentuma van. (lasd Command struktura.)
   Visszateres utan which tarolja, hogy a cmds tomb hanyadik eleme
   a beolvasott parancs, es parnum, hogy hany argumentuma van.
   (ez akkor kell, ha a Command struktura valtozo argumentumszamot hataroz meg)
   cmds tomb utolso eleme NULL kell hogy legyen.
   
   a visszaadott GList felszabaditasa a hivo fv felelossege.
*/
GList *read_command(SSL *ssl, Command *cmds, int *which, int *parnum)
{
	GList *line,*p;
	int cnt,i;
	
	line=read_token_line(ssl);
	if(!line){
		*which=-1;
		*parnum=0;
		return NULL;
	}
	
	cnt=0;
	p=line;
	while(p){
		cnt++;
		p=p->next;
	}
	cnt--;
	
	if(cnt<0){    // bar ha jobban vegiggondoljuk, ez lehetetlen
		*which=-1;
		*parnum=0;
		free_g_list_of_strings(line);
		return NULL;
	}
	
	i=0;
	while(cmds[i].cmd){
		if( (cnt==cmds[i].argnum || cmds[i].argnum==-1) && 
		!strcmp(cmds[i].cmd,line->data)){
			*which=i;
			*parnum=cnt;
			return line;
		}
		i++;
	}
	
	*which=-1;
	*parnum=0;
	free_g_list_of_strings(line);
	return NULL;
}

/*
   visszaadja, hogy kell-e vedeni egy karaktert.
*/
static int is_protected(char a)
{
	if(a=='\\') return 1;
	if(a==' ') return 1;
	if(a=='\n') return 1;
	return 0;
}

/*
   beleir az atadott SSL connectionba egy sort. A cnt parameter hatarozza meg,
   hogy hany token van osszesen parameterkent atadva a fv-nek.
   Ezeket egymastol szokozzel elvalasztva irja ki a fv, backslash-sel vedve
   a vedendo karaktereket. (pl space);
   
   a sor max hossza BUFFER_LEN byte;
*/	
int write_token_line(SSL *ssl,char *fmt,...)
{
	char buffer[BUFFER_LEN],*current;
	GList *list;
	char **array;
	int i,fcnt,gcnt,lcnt,vedett;
	va_list args;
	
	va_start(args,fmt);
	
	fcnt=0;
	gcnt=0;
	while(fmt[fcnt]){
		switch(fmt[fcnt]){
		case 's':
			lcnt=0;
			current=va_arg(args,char *);
			while(current[lcnt]){
				if(is_protected(current[lcnt]) && !vedett){
					buffer[gcnt++]='\\';
					vedett=1;
				}else{
					buffer[gcnt++]=current[lcnt++];
					vedett=0;
				}
				if(gcnt==BUFFER_LEN-2){ // -s, mert lezaro \n, es lezaro 0
					va_end(args);       // bar a lezaro 0 valojaban felesleges
					return 0;
				}
			}
			if(fmt[fcnt+1]) buffer[gcnt++]=' ';
			else buffer[gcnt++]='\n';
			break;
		case 'g':
			list=va_arg(args,GList *);
			while(list){
				lcnt=0;
				current=list->data;
				while(current[lcnt]){
					if(is_protected(current[lcnt]) && !vedett){
						buffer[gcnt++]='\\';
						vedett=1;
					}else{
						buffer[gcnt++]=current[lcnt++];
						vedett=0;
					}
					if(gcnt==BUFFER_LEN-2){ // -s, mert lezaro \n, es lezaro 0
						va_end(args);       // bar a lezaro 0 valojaban felesleges
						return 0;
					}
				}
				if(list->next)buffer[gcnt++]=' ';
				else if(fmt[fcnt+1])buffer[gcnt++]=' ';
				else buffer[gcnt++]='\n';
				list=list->next;
			}
			break;
		case 'a':
			i=0;
			array=va_arg(args,char **);
			while(array[i]){
				lcnt=0;
				current=array[i];
				while(current[lcnt]){
					if(is_protected(current[lcnt]) && !vedett){
						buffer[gcnt++]='\\';
						vedett=1;
					}else{
						buffer[gcnt++]=current[lcnt++];
						vedett=0;
					}
					if(gcnt==BUFFER_LEN-2){ // -2, mert lezaro \n, es lezaro 0
						va_end(args);       // bar a lezaro 0 valojaban felesleges
						return 0;
					}
				}
				if(array[i+1])buffer[gcnt++]=' ';
				else if(fmt[fcnt+1])buffer[gcnt++]=' ';
				else buffer[gcnt++]='\n';
				i++;
			}
			break;
		default:
		}
		fcnt++;
	}
	buffer[gcnt]=0;

	va_end(args);

#if 0	
#ifdef Z_CURSES	
	printw("%s\n",buffer);
#else
	printf("%s\n",buffer);
#endif
#endif

	return SSL_write(ssl,buffer,gcnt);
}
