/***************************************************************************
 *
 * Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: zpython.h,v 1.13 2000/08/08 15:00:51 bazsi Exp $
 *
 ***************************************************************************/

#ifndef ZORP_ZPYTHON_H_INCLUDED
#define ZORP_ZPYTHON_H_INCLUDED

#include <Python.h>
#include <zorp/sockaddr.h>
#include <zorp/stream.h>
#include <zorp/io.h>
#include <zorp/labelset.h>
#include <zorp/authconn.h>
#include <zorp/satyr.h>

/*+

  ZPolicyObj is a reference counted data structure, capable of holding
  a value in the policy layer. Currently it's just a simple PyObject.

  +*/
typedef PyObject ZPolicyObj;

/*+

  Each C thread needs a Policy thread data structure.

  +*/
typedef PyThreadState ZPolicyThread;

extern PyThreadState *z_python_global_state;

gboolean z_python_init(void);
gboolean z_python_destroy(void);

void z_python_lock(void);
void z_python_unlock(void);

/*+

  ZorpStream is the Python interface to ZStream.

  +*/
typedef struct _ZorpStream
{
  PyObject_HEAD
  ZStream *stream;
} ZorpStream;

extern PyTypeObject z_py_zorp_stream_type;

#define z_py_zorp_stream_check(ob) ((ob)->ob_type == &z_py_zorp_stream_type)

void z_py_zorp_stream_init(void);
void z_py_zorp_proxy_init(void);

/* pylisten */
/*+

  ZorpListen is the python interface to ZIOListen

  +*/
typedef struct _ZorpListen ZorpListen;

void z_py_zorp_listen_init(void);

/* pyreceive */
void z_py_zorp_receive_init(void);

/* pyconnect */

/*+
  ZorpConnect is the Python interface to ZIOConnect.
  +*/
typedef struct _ZorpConnect ZorpConnect;

void z_py_zorp_connect_init(void);

/* pyproxy */

struct _ZProxy;

/*+
  ZorpProxy is the Python interface to ZProxy.
  +*/
typedef struct _ZorpProxy
{
  PyObject_HEAD
  PyObject *name;
  struct _ZProxy *proxy;
} ZorpProxy;

/* pysockaddr */

/*+

  ZorpSockAddr is the Python interface to ZSockAddr

  +*/
typedef struct _ZorpSockAddr
{
  PyObject_HEAD
  ZSockAddr *sa;
} ZorpSockAddr;

extern PyTypeObject z_py_zorp_sockaddr_type;

PyObject *z_py_zorp_sockaddr_new(ZSockAddr *sa);
#define z_py_zorp_sockaddr_check(ob) ((ob)->ob_type == &z_py_zorp_sockaddr_type)

void z_py_zorp_sockaddr_init(void);

/*+

  ZorpLabelSet is the Python interface to ZLabelSet. It's not directly used by
  the native C part of Zorp, only exported to Python to do something with it.

  +*/

typedef struct _ZorpLabelSet
{
  PyObject_HEAD
  ZLabelSet *labelset;
} ZorpLabelSet;

extern PyTypeObject z_py_zorp_label_set_type;

PyObject *z_py_zorp_labelset_new(ZLabelSet *labelset);
#define z_py_zorp_labelset_check(ob) ((ob)->ob_type == &z_py_zorp_sockaddr_type)

/* void z_py_zorp_label_set_init(void); */

/*+

  Wrapper class for ZAuthConnection.

  +*/

typedef struct _ZorpAuthConnection 
{
  PyObject_HEAD
  ZAuthConnection *authconn;
} ZorpAuthConnection;

extern PyTypeObject z_py_zorp_authconn_type;

#define z_py_zorp_authconn_check(ob) ((ob)->ob_type == &z_py_zorp_authconn_type)

/* PyObject *z_py_zorp_auth_connection_new(ZAuthConnection *authconn); */

void z_py_zorp_authconn_init(void);

/*+

  Wrapper class for ZSatyrAuthenticator.

  +*/

typedef struct _ZorpSatyrAuthenticator 
{
  PyObject_HEAD
  ZSatyrAuthenticator *satyr;
} ZorpSatyrAuthenticator;

extern PyTypeObject z_py_zorp_satyr_authenticator_type;

/* PyObject *z_py_zorp_satyr_authenticator_new(ZAuthConnection *authconn); */

void z_py_zorp_satyr_authenticator_init(void);


#endif
