/***************************************************************************
 *
 * Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: ftp.c,v 1.18 2000/08/08 17:19:02 bazsi Exp $
 *
 * Author:  Andras Kis-Szabo <kisza@sch.bme.hu>
 * Auditor:
 * Notes:
 *
 ***************************************************************************/

#include <zorp/zorp.h>
#include <zorp/registry.h>
#include <zorp/stream.h>
#include <zorp/readline.h>
#include <zorp/proxy.h>
#include <zorp/policy.h>
#include <zorp/thread.h>
#include <zorp/zpython.h>
#include <zorp/log.h>
#include <zorp/sockaddr.h>

#include <sys/socket.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <errno.h>


#include "ftp.h"
#include "ftphash.h"
#include "ftpmain.h"
#include "ftpparse.h"

GHashTable *ftp_answer_hash = NULL;
GHashTable *ftp_command_hash = NULL;

//#define ENABLE_DEBUG
void ftp_proxy_free (ZProxy * s);

// Function:    ftp_read_line_get
// In:          FtpProxy*               proxy pointer
//              gchar                   S/C
//              char *                  line
//              int                     line length
//              int                     telnet seq len
// Out:         int                     from z_read_line_get
// Desc:        TELNET sequence handling
//      IACl - needed?
//      Only if we want pass the sequences to the upper layers
int
ftp_read_line_get (FtpProxy * self, gchar side, char **line, int *length, int *IACl)
{
  int readback;
  int i;
  int state;
  char *tmp;
  char *tmp2;
  int pos1;
  int pos2;
  int iaciac = 0;
  unsigned char funcs[10];
  unsigned char negot[5];
#if 0
  static char *line = NULL;
  int length = MAX_LEN;
#endif

  *IACl = 0;
  pos1 = 0;
  pos2 = 0;
  funcs[0] = 241;
  funcs[1] = 242;
  funcs[2] = 243;
  funcs[3] = 244;
  funcs[4] = 245;
  funcs[5] = 246;
  funcs[6] = 247;
  funcs[7] = 248;
  funcs[8] = 249;
  funcs[9] = '\0';
  negot[0] = 251;
  negot[1] = 252;
  negot[2] = 253;
  negot[3] = 254;
  negot[4] = '\0';

#if 0
  if (line == NULL)
    {
      if ((line = g_new0 (char, length)) == NULL)
	{
	  // quit
	}
    }
#endif

  switch (side)
    {
    case 'S':
      readback = z_read_line_get (self->readline[EP_SERVER], line, length);
      break;
    case 'C':
      readback = z_read_line_get (self->readline[EP_CLIENT], line, length);
      break;
    default:
      break;
    }

  if (*line == 0)
    return readback;
  if (*line == NULL)
    return readback;

  tmp = g_new0 (char, (*length) + 2);
  tmp2 = g_new0 (char, (*length) + 3);

  state = FTP_TELNET;

  // IAC IP 
  // IAC DO ACK
  if (*length)
    for (i = 0; i < *length; i++)
      {
	// TODO: adding SB SE
	switch (state)
	  {
	  case FTP_TELNET:
	    if ((unsigned char) (*line)[i] != 255)
	      {
		tmp[pos1] = (*line)[i];
		pos1++;
	      }
	    else
	      {
		if ((unsigned char) (*line)[i + 1] == 255)
		  {
		    tmp[pos1] = (*line)[i];
		    pos1++;
		    iaciac++;
		    i++;
		  }
		else
		  {
		    state = FTP_TELNET_IAC;
		    tmp2[pos2] = (*line)[i];
		    pos2++;
		  }
	      }
	    break;
	  case FTP_TELNET_IAC:
	    if (strchr (funcs, (*line)[i]))
	      {
		// in funcs
		state = FTP_TELNET;
		tmp2[pos2] = (*line)[i];
		pos2++;
	      }
	    else if (strchr (negot, (*line)[i]))
	      {
		// in negotiation
		state = FTP_TELNET_IAC_DW;
		switch ((unsigned char) (*line)[i])
		  {
		  case 251:	// WILL
		    tmp2[pos2] = 252;
		    pos2++;
		    break;
		  case 252:	// WONT
		    tmp2[pos2] = 252;
		    pos2++;
		    break;
		  case 253:	// DO
		    tmp2[pos2] = 254;
		    pos2++;
		    break;
		  case 254:	// DONT
		    tmp2[pos2] = 254;
		    pos2++;
		    break;
		  default:
		    // TODO bad telnet seq
		    break;
		  }
	      }
	    else if ((unsigned char) (*line)[i] == 250)
	      {
		state = FTP_TELNET_IAC_FUNC;
		tmp2[pos2] = (*line)[i];
		pos2++;
	      }
	    else
	      {
		// Bad seq
	      }
	    break;
	  case FTP_TELNET_IAC_DW:
	    state = FTP_TELNET;
	    tmp2[pos2] = (*line)[i];
	    pos2++;
	    break;
	  case FTP_TELNET_IAC_FUNC:
	    if ((unsigned char) (*line)[i] == 240)
	      {
		state = FTP_TELNET;
	      }
	    tmp2[pos2] = (*line)[i];
	    pos2++;
	    break;
	  default:
	    break;
	  }
      }

  tmp2[pos2] = '\0';
  *IACl = strlen (tmp2);
  (*length) -= (*IACl);
  (*length) -= iaciac;
  snprintf (*line, *length + 1, "%s", tmp);
  strncat (*line, tmp2, *IACl);

  if (strlen (tmp2) > 0)
    switch (side)
      {
      case 'S':
	z_stream_write (self->super.endpoints[EP_CLIENT], tmp2, pos2, IACl);
	break;
      case 'C':
	z_stream_write (self->super.endpoints[EP_SERVER], tmp2, pos2, IACl);
	break;
      default:
	break;
      }
#ifdef ENABLE_DEBUG
  z_debug (7, "%s: telnet answer -%s- ", self->super.session_id, tmp2);
#endif

  if (tmp)
    g_free (tmp);
  if (tmp2)
    g_free (tmp2);

  // TODO: bad telnet sequence
  return readback;
}

// Function:    ftp_init_stream_server
// In:          FtpProxy*               proxy pointer
// Out:         -
// Desc:        create streams
void
ftp_init_stream_server (FtpProxy * self)
{
  ZorpSockAddr *tmpsa;
  gchar *tmp;

  tmp = g_new0 (gchar, 17);

  if (!self->super.endpoints[EP_SERVER])
    {
      self->quit = 1;
      return;
    }
  self->readline[EP_SERVER] = z_read_line_new (self->super.endpoints[EP_SERVER], MAX_LEN, ZRL_EOL_CRLF | ZRL_TRUNCATE);
  
  z_policy_lock(self->super.thread);
  tmpsa = (ZorpSockAddr *)z_session_getattr(self->super.handler, "server_addr");
  z_policy_unlock(self->super.thread);
  if (tmpsa && z_py_zorp_sockaddr_check(tmpsa))
    {
      z_inet_ntoa (tmp, 16, ((struct sockaddr_in *) &tmpsa->sa->sa)->sin_addr);
      g_string_assign (self->server_data_address, tmp);
      self->server_data_port = ntohs (((struct sockaddr_in *) &tmpsa->sa->sa)->sin_port) - 1;
#ifdef ENABLE_DEBUG
      z_debug (7, "%s: ---- server %s %d", self->super.session_id, self->server_data_address->str, self->server_data_port);
#endif
      z_policy_var_unref(tmpsa);
    }
  else
    {
       z_warning(0, "%s: internal error, no session.server_addr defined", self->super.session_id);
    }

  ftp_test(self);

  g_free (tmp);

}

// Function:    ftp_init_stream_client
// In:          FtpProxy*               proxy pointer
// Out:         -
// Desc:        create streams
void
ftp_init_stream_client (FtpProxy * self)
{
  gchar *tmp;
  ZorpSockAddr *tmpsa;

  tmp = g_new0 (gchar, 17);

  if (!self->super.endpoints[EP_CLIENT])
    {
      self->quit = 1;
      return;
    }
  self->readline[EP_CLIENT] = z_read_line_new (self->super.endpoints[EP_CLIENT], MAX_LEN, ZRL_EOL_CRLF | ZRL_TRUNCATE);

  z_policy_lock(self->super.thread);
  tmpsa = (ZorpSockAddr *)z_session_getattr(self->super.handler, "client_addr");
  z_policy_unlock(self->super.thread);
  if (tmpsa && z_py_zorp_sockaddr_check(tmpsa))
    {
      z_inet_ntoa (tmp, 16, ((struct sockaddr_in *) &tmpsa->sa->sa)->sin_addr);
      g_string_assign (self->client_data_address, tmp);
      self->client_data_port = ntohs (((struct sockaddr_in *) &tmpsa->sa->sa)->sin_port);
#ifdef ENABLE_DEBUG
      z_debug (7, "%s: ---- client %s %d", self->super.session_id, self->client_data_address->str, self->client_data_port);
#endif
    }
  else
    {
      z_warning(0, "%s: internal error, no session.server_addr defined", self->super.session_id);
    }

  ftp_test(self);

  g_free (tmp);

}

// Function:    ftp_227PASV
// In:          FtpProxy                proxy pointer
// Out:         int                       
// Desc:        227 PASV answer parser
int
ftp_227PASV (FtpProxy * self, char *line, int length, char *newip)
{
  //gchar *line;
  //guint length = 0;
  //guint IACl = 0;
  //int parse;
  //int release = 0;
  int i;
  //guint bytes_written;
  //int readback;
  //char ansh[11];
  //char newip[31];
  int newport;
  char *sp;
  char *buf;
  unsigned int fakeport;
  //ZPolicyObj *res;
  FTPCommandDescriptor *workp;

#ifdef ENABLE_DEBUG
  z_debug (7, "%s: answer conv PASV", self->super.session_id);
#endif

  ftp_test(self);

  self->last_port = self->tmp_port;

  buf = g_new0 (char, length + 2);
  snprintf (buf, length + 1, "%.*s", (int) length, line);

#ifdef ENABLE_DEBUG
  z_debug (7, "%s: interesting ... %s", self->super.session_id, buf);
#endif

  if (
      (strrchr (buf, ')') < strrchr (buf, '(')) ||
      (strrchr (buf, ')') - buf > length) || (strrchr (buf, '(') - buf > length) || (strrchr (buf, ')') - strrchr (buf, '(') > 26))

    {
      // TODO: drop
#ifdef ENABLE_DEBUG
      z_debug (7, "%s: bad answer - need drop", self->super.session_id);
#endif
      return 1;
    }
  else
    {
#ifdef ENABLE_DEBUG
      z_debug (7, "%s: he? -%s-", self->super.session_id, buf);
#endif
      strncpy (newip, strrchr (buf, '(') + 1, strrchr (buf, ')') - strrchr (buf, '(') - 1);
      newip[strrchr (buf, ')') - strrchr (buf, '(') - 1] = '\0';

#ifdef ENABLE_DEBUG
      z_debug (7, "%s: %s %d", self->super.session_id, newip, strlen (newip));
#endif
      if (strtol (newip, NULL, 10) > 255)
	{
	  return 1;
	}
      if ( errno == ERANGE )
        {
	  return 1;
	}
      for (i = 0; i < strlen (newip); i++)
	{
	  if (!isdigit (newip[i]))
	    {
	      if (newip[i] == ',')
		{
		  if (strtol (newip + i + 1, NULL, 10) > 255)
		    {
		      return 1;
		    }
                  if ( errno == ERANGE )
                    {
	              return 1;
	            }
		  newip[i] = '.';
		}
	      else
		{
		  return 1;
		}
	    }
	}

      sp = strrchr (newip, '.');
      *sp = ',';
      sp = strrchr (newip, '.');
      *sp = '\0';
      sp++;
      newport = strtol (sp, NULL, 10);
      newport *= 256;
      sp = strrchr (sp, ',');
      sp++;
      newport += strtol (sp, NULL, 10);
#ifdef ENABLE_DEBUG
      z_debug (7, "%s: %s %d", self->super.session_id, newip, newport);
#endif

      z_policy_lock (self->super.thread);
      i = z_policy_event (self->super.handler, "pasvAnswer", z_policy_var_build ("cs#si", 'C', line, length, newip, newport));
      z_policy_unlock (self->super.thread);

      switch (i)
	{
	case Z_ACCEPT:
	  switch (self->client_portpasv)
	    {
	    case 0:
	    case 2:
	      if ( (self->client_portpasv == 2) && (self->last_port) )
	        {
		  // backup
		  g_string_assign (self->client_data_address, self->tmp_data_address->str);
		  self->client_data_port = self->tmp_data_port;
	        }
	      if (self->bounce_check)
		{
		  if (strcmp (self->server_data_address->str, newip))
		    {
		      // BOUNCED
#ifdef ENABLE_DEBUG
		      z_debug (7, "%s: BOUNCE!!!!", self->super.session_id);
#endif
		      // TODO: msgs[MSG_SERVER_BOUNCE]
		      g_string_assign (self->server_data_address, newip);
		      self->server_data_port = newport;

		      if ( (self->internal_server) && (self->NAT) && (!self->last_port) )
		        {
		          if ((fakeport=ftp_get_port(self,self->fw_external_data_address->str,0)))
			    {
			      self->fw_external_data_port = fakeport;
			      snprintf (newip, 30, "227 (%s,%d,%d)", self->fw_external_data_address->str, self->fw_external_data_port / 256, self->fw_external_data_port % 256);
		              g_strdelimit (newip, ".", ',');
			    }
		          else
			    {
			      // TODO: unsuccessful listen
			      return 1;    
			    }
			}
		    }
		  else
		    {
		      g_string_assign (self->server_data_address, newip);
		      self->server_data_port = newport;

		      if ( (self->internal_server) && (self->NAT) && (!self->last_port) )
		        {
		          if ((fakeport=ftp_get_port(self,self->fw_external_data_address->str,0)))
			    {
			      self->fw_external_data_port = fakeport;
			      snprintf (newip, 30, "227 (%s,%d,%d)", self->fw_external_data_address->str, self->fw_external_data_port / 256, self->fw_external_data_port % 256);
		              g_strdelimit (newip, ".", ',');
			    }
		          else
			    {
			      // TODO: unsuccessful listen
			      return 1;    
			    }
			}

		    }
		}
	      if ((workp = (FTPCommandDescriptor *) ftp_command_hash_get ("PASV")) == NULL)
	        {
	        }
	      else
	        {
		  ftp_data_channel(self,workp);
	        }
	      break;
	    case 1:
	        // TODO impossible
	      break;
	    default:
	      break;
	    }

	  break;
	case Z_DENY:
	  break;
	case Z_UNSPEC:
	case Z_REJECT:
	  break;
	case Z_ABORT:
	  ftp_command_drop (self, line, length, msgs[MSG_Z_ABORT]);
	  self->quit = 1;
	  self->state = FTP_MAIN_RET;
	  return 1;
	  break;
	}

    }

  if (buf)
    g_free (buf);

  ftp_test(self);

  return 0;

}

// Function:    ftp_200PORT
// In:          FtpProxy                proxy pointer
// Out:         int                       
// Desc:        200 PORT answer parser
int
ftp_200PORT (FtpProxy * self, char *line, int length, char *newip)
{
  unsigned int fakeport;
  FTPCommandDescriptor *workp;

  ftp_test(self);

  self->last_port = self->tmp_port;

  switch (self->client_portpasv)
    {
    case 0:
      // PORT enabled
      if ((strcmp (self->client_data_address->str, self->tmp_data_address->str)) && (self->bounce_check))
	{
	  // BOUNCED
#ifdef ENABLE_DEBUG
	  z_debug (7, "%s: BOUNCE!!!!", self->super.session_id);
#endif
	  // TODO: msgs[MSG_SERVER_BOUNCE]
	  g_string_assign (self->client_data_address, self->tmp_data_address->str);
	  self->client_data_port = self->tmp_data_port;
	  if ( (!self->internal_server) && (self->NAT) )
	    {
	      self->fw_external_data_port = self->tmp2_data_port;
	    }
	}
      else
	{
	  g_string_assign (self->client_data_address, self->tmp_data_address->str);
	  self->client_data_port = self->tmp_data_port;
	  if ( (!self->internal_server) && (self->NAT) )
	    {
	      self->fw_external_data_port = self->tmp2_data_port;
	    }
	}
      break;
    case 1:
      if (!self->last_port)
	{
	  // PASV -> PORT
	  if (!self->NAT)
	    {
	      if ((fakeport=ftp_get_port(self,self->server_data_address->str,1)))
		{
		  self->server_data_port = fakeport;
		  snprintf (newip, 30, "227 (%s,%d,%d)", self->server_data_address->str, self->server_data_port / 256, self->server_data_port % 256);
		}
	      else
		{
		  // TODO: unsuccessful listen
		}
	    }
	  else
	    {
	      if (self->internal_server)
	        {
	          if ((fakeport=ftp_get_port(self,self->fw_external_data_address->str,0)))
		    {
		      self->fw_external_data_port = fakeport;
		      snprintf (newip, 30, "227 (%s,%d,%d)", self->fw_external_data_address->str, self->fw_external_data_port / 256, self->fw_external_data_port % 256);
		    }
	          else
		    {
		      // TODO: unsuccessful listen
		    }
		}
	      else
	        {
	          if ((fakeport=ftp_get_port(self,self->server_data_address->str, 1)))
		    {
		      self->server_data_port = fakeport;
		      snprintf (newip, 30, "227 (%s,%d,%d)", self->server_data_address->str, self->server_data_port / 256, self->server_data_port % 256);
		    }
	          else
		    {
		      // TODO: unsuccessful listen
		    }
	        }
	    }

	  g_strdelimit (newip, ".", ',');
	}
      else
	{
          if ((strcmp (self->client_data_address->str, self->tmp_data_address->str)) && (self->bounce_check))
            {
              // BOUNCED
#ifdef ENABLE_DEBUG
              z_debug (7, "%s: BOUNCE!!!!", self->super.session_id);
#endif
              // TODO: msgs[MSG_SERVER_BOUNCE]
              g_string_assign (self->client_data_address, self->tmp_data_address->str);
              self->client_data_port = self->tmp_data_port;
              if ( (!self->internal_server) && (self->NAT) )
                {
                  self->fw_external_data_port = self->tmp2_data_port;
                }
            }
          else
            {
              g_string_assign (self->client_data_address, self->tmp_data_address->str);
              self->client_data_port = self->tmp_data_port;
              if ( (!self->internal_server) && (self->NAT) )
                {
                  self->fw_external_data_port = self->tmp2_data_port;
                }
            }
	}
      break;
    case 2:
       // TODO: impossible
      break;
    default:
      break;

    }

	if ((workp = (FTPCommandDescriptor *) ftp_command_hash_get ("PORT")) == NULL)
	  {
	  }
	else
	  {
	    ftp_data_channel(self,workp);
	  }

  ftp_test(self);

  return 0;
}

// Function:    ftp_thread_server
// In:          FtpProxy                proxy pointer
//              int                     !send back the msg?
// Out:         int                       
// Desc:        ftp proxy thread: server -> proxy 
int
ftp_thread_server (FtpProxy * self, int nont)
{
  gchar *line;
  guint length = 0;
  guint IACl = 0;
  int parse;
  int release = 0;
  guint bytes_written;
  int readback;
  char ansh[11];
  char newip[31];

  //
  // FROM server TO client
  //  
#ifdef ENABLE_DEBUG
  z_debug (7, "%s: S -> C", self->super.session_id);
#endif
  release = 0;
  parse = PARSE_SERVER_BEGIN;
  // 
  // server answer parser
  // 
  while (!release)
    {
#ifdef ENABLE_DEBBUG
      z_debug (7, "%s: readline elott", self->super.session_id);
#endif
      readback = ftp_read_line_get (self, 'S', &line, &length, &IACl);

#ifdef ENABLE_DEBUG
      z_debug (7, "%s: get %d %.*s", self->super.session_id, length, (int) length, line);
#endif

      if (length == 0)
	{
	  release = 1;
	  self->state = FTP_MAIN_RET;
	  break;
	}

#ifdef ENABLE_DEBUG
      z_debug (7, "%s: delimer %d %d %d %d", self->super.session_id, line[length - 4], line[length - 3], line[length - 2], line[length - 1]);
#endif
      if (ST_ERROR (readback) != ST_OK)
	{
#ifdef ENABLE_DEBUG
	  z_debug (7, "%s: bad delimer! \\r\\n", self->super.session_id);
#endif
	  if ((self->log_answer) && (self->log_dropped))
	    {
	      // Try until unsuccessful...
	      while (!ftp_log (self, LOG_ANSW + LOG_DROP, line, length));
	    }
	  release = 1;
	  self->state = FTP_MAIN_RET;
	  break;
	}

      // Answer states, depends on lines...
      // release: correct closing message "??? "
      switch (parse)
	{
	case PARSE_SERVER_BEGIN:
	  // New answer ...
#ifdef ENABLE_DEBUG
	  z_debug (7, "%s: PARSE_SERVER_BEGIN", self->super.session_id);
#endif
	  // get the code from the 1st line
	  // wrong format -> drop the connection
	  if ((!isdigit ((int) line[0])) || (!isdigit ((int) line[1])) || (!isdigit ((int) line[2])) || ((line[3] != ' ') && (line[3] != '-')))
	    {
#ifdef ENABLE_DEBUG
	      z_debug (7, "%s: FTP: bad answerform at begining!", self->super.session_id);
#endif
	      if ((self->log_answer) && (self->log_dropped))
		{
		  while (!ftp_log (self, LOG_ANSW + LOG_DROP, line, length));
		}
	      parse = PARSE_RETURN;
	      self->state = FTP_MAIN_RET;
	      break;
	    }
	  strncpy (self->answer_code, line, 3);
#ifdef ENABLE_DEBUG
	  z_debug (7, "%s: answer code %3s", self->super.session_id, self->answer_code);
#endif

	  // 
	  // acceptable answers from the server
	  // 
#ifdef ENABLE_DEBUG
	  z_debug (7, "%s:  - ftp_state - %d - ", self->super.session_id, self->ftp_state);
#endif
	  snprintf (ansh, 11, "%.2d%3s%s", self->ftp_state, self->answer_code, self->command);

#ifdef ENABLE_DEBUG
	  z_debug (7, "%s:  - ANSWER_HASH - %s - ", self->super.session_id, ansh);
#endif
	  if ((!ftp_answer_hash_get (ansh)) && (!ftp_local_answer_hash_get (self, ansh)))
	    {
#ifdef ENABLE_DEBUG
	      z_debug (7, "%s: FTP: bad answer %d", self->super.session_id, ftp_local_answer_hash_get (self, ansh));
#endif
	      if ((self->log_answer) && (self->log_dropped))
		{
		  while (!ftp_log (self, LOG_ANSW + LOG_DROP, line, length));
		}
	      self->ftp_state = FTP_STATE_RETURN;
	      release = 1;
	      parse = PARSE_RETURN;
	      self->state = FTP_MAIN_RET;
	      break;
	    }
	  else
	    {
#ifdef ENABLE_DEBUG
	      z_debug (7, "%s: answer", self->super.session_id);
#endif
	      switch (self->ftp_state)
		{
		case FTP_STATE_CONNECT:
		  break;
		case FTP_STATE_LOGINAUTH:
		case FTP_STATE_LOGIN:
		  break;
		case FTP_STATE_LOGIN_U:
		  break;
		case FTP_STATE_LOGIN_P:
		  // 230 after pass -> go to convers. state
		  if (!strncmp (self->answer_code, "230", 3))
		    {
		      self->ftp_state = FTP_STATE_CONVERSATION;
		    }
		  break;
		case FTP_STATE_LOGIN_A:
		  break;
		case FTP_STATE_CONVERSATION:
#ifdef ENABLE_DEBUG
		  z_debug (7, "%s:  answer conv.", self->super.session_id);
#endif
		  if (!strncmp (ansh, "04227PASV", strlen (ansh)))
		    {
		      if (ftp_227PASV (self, line, length, newip))
			{
			  self->ftp_state = FTP_STATE_RETURN;
			  release = 1;
			  parse = PARSE_RETURN;
			  self->state = FTP_MAIN_RET;
			  break;
			}
		    }
		  else
		    {
		      // TODO: answer event
		      if (!strncmp (ansh, "04200PORT", strlen (ansh)))
			{
			  if(ftp_200PORT (self, line, length, newip))
			   {
			     self->ftp_state = FTP_STATE_RETURN;
			     release = 1;
			     parse = PARSE_RETURN;
			     self->state = FTP_MAIN_RET;
			     break;
			   }
			}
		    }
		  break;
		case FTP_STATE_QUIT:
		  break;
		case FTP_STATE_RETURN:
		  parse = PARSE_RETURN;
		  break;
		default:
		  break;
		}
	    }

	  // Cut the oversized answer
	  if (length >= self->max_line_length)
	    {
#ifdef ENABLE_DEBUG
	      z_debug (7, "%s: Csonkolok!", self->super.session_id);
#endif
	      length = self->max_line_length;
	    }

	  // All ok, server answer accepted,
	  // new state: in the answer
	  parse = PARSE_SERVER_INNER;

	  // If the  PORT/PASV translation is active,
	  // the answer is performed by the translator
	  // if we need log the answer, do it!
	  if (self->verbose_messages)
	    {
	      // PASV -> PORT
	      if ((!strncmp (ansh, "04200PORT", strlen (ansh))) && (self->client_portpasv == 1) && (!self->last_port))
		{
		  ftp_stream_write (self, 'C', newip, strlen (newip));
		  if (self->log_answer)
		    {
		      while (!ftp_log (self, LOG_ANSW + LOG_PASS + LOG_PROX, newip, strlen (newip)));
		    }
		}
	      else
		// PORT -> PASV
	      if ((!strncmp (ansh, "04227PASV", strlen (ansh))) && (self->client_portpasv == 2) && (self->last_port))
		{
		  ftp_stream_write (self, 'C', msgs[MSG_PORT_SUCCESS], strlen (msgs[MSG_PORT_SUCCESS]));
		  if (self->log_answer)
		    {
		      while (!ftp_log (self, LOG_ANSW + LOG_PASS + LOG_PROX, msgs[MSG_PORT_SUCCESS], strlen (msgs[MSG_PORT_SUCCESS])));
		    }

		}
	      else
		{
		  // linear
		  if ((((!strncmp (ansh, "04227PASV", strlen (ansh))) && (self->client_portpasv == 0) && (!self->last_port)) ||
		       ((!strncmp (ansh, "04227PASV", strlen (ansh))) && (self->client_portpasv == 2) && (!self->last_port))) && (self->NAT) && (self->internal_server))
		    {
		      // NAT, internal server & PASV answer
		      ftp_stream_write (self, 'C', newip, strlen (newip));
		      if (self->log_answer)
			{
			  while (!ftp_log (self, LOG_ANSW + LOG_PASS + LOG_PROX, newip, strlen (newip)));
			}
		    }
		  else
		    {
		      // OK
		      if (!nont)
			ftp_stream_write (self, 'C', line, length);
		    }
		}
	      if (self->log_answer)
		{
		  while (!ftp_log (self, LOG_ANSW + LOG_PASS, line, length));
		}

	      if (line[3] == ' ')
		{
		  release = 1;
		  ftp_answerstate (self);
		  parse = PARSE_SERVER_BEGIN;
		}
	    }

	  // if we needn't log the answer, and this is
	  // the last line, log the sort answer
	  if ((!self->verbose_messages) && (line[3] == ' '))
	    {
	      // PASV -> PORT
	      if ((!strncmp (ansh, "04200PORT", strlen (ansh))) && (self->client_portpasv == 1) && (!self->last_port))
		{
		  ftp_stream_write (self, 'C', newip, strlen (newip));
		  if (self->log_answer)
		    {
		      while (!ftp_log (self, LOG_ANSW + LOG_PASS + LOG_PROX, newip, strlen (newip)));
		    }
		}
	      else
		// PORT -> PASV
	      if ((!strncmp (ansh, "04227PASV", strlen (ansh))) && (self->client_portpasv == 2) && (self->last_port))
		{
		  ftp_stream_write (self, 'C', "200 ", strlen ("200 "));
		  if (self->log_answer)
		    {
		      while (!ftp_log (self, LOG_ANSW + LOG_PASS + LOG_PROX, msgs[MSG_PORT_SUCCESS], strlen (msgs[MSG_PORT_SUCCESS])));
		    }

		}
	      else
		{
		  // linear
		  if ((((!strncmp (ansh, "04227PASV", strlen (ansh))) && (self->client_portpasv == 0) && (!self->last_port)) ||
		       ((!strncmp (ansh, "04227PASV", strlen (ansh))) && (self->client_portpasv == 2) && (!self->last_port))) && (self->NAT) && (self->internal_server))
		    {
		      // NAT, internal server & PASV answer
		      ftp_stream_write (self, 'C', newip, strlen (newip));
		      if (self->log_answer)
			{
			  while (!ftp_log (self, LOG_ANSW + LOG_PASS + LOG_PROX, newip, strlen (newip)));
			}
		    }
		  else
		    {
		      // OK
		      if (!nont)
			ftp_stream_write (self, 'C', line, 4);
		    }
		}
	      if (self->log_answer)
		{
		  while (!ftp_log (self, LOG_ANSW + LOG_PASS, line, length));
		}
	      release = 1;
	      ftp_answerstate (self);
	      parse = PARSE_SERVER_BEGIN;
	    }
	  break;

	case PARSE_SERVER_INNER:
	  // We are in a server answer sequence
#ifdef ENABLE_DEBUG
	  z_debug (7, "%s: PARSE_SERVER_INNER", self->super.session_id);
#endif
	  // begin with numbers?
	  if (((isdigit ((int) line[0])) && (isdigit ((int) line[1])) && (isdigit ((int) line[2])) && ((line[3] == '-') || (line[3] == ' ')))
	      || (!isdigit ((int) line[0])))
	    {
	      // wrong sequence?
	      if ((isdigit ((int) line[0])) && (isdigit ((int) line[1])) && (isdigit ((int) line[2])) && (strncmp (self->answer_code, line, 3)))
		{
		  // Wrong answer!!!
#ifdef ENABLE_DEBUG
		  z_debug (7, "%s: FTP: break in answer sequence!", self->super.session_id);
#endif
		  if ((self->log_answer) && (self->log_dropped))
		    {
		      while (!ftp_log (self, LOG_ANSW + LOG_DROP, line, length));
		    }
		  parse = PARSE_RETURN;
		  self->state = FTP_MAIN_RET;
		  break;
		}
	      else
		{
#ifdef ENABLE_DEBUG
		  z_debug (7,
			   "%s: Sequence looks good or !isdigit([0]) %c%c%c - %c%c%c", self->super.session_id,
			   line[0], line[1], line[2], self->answer_code[0], self->answer_code[1], self->answer_code[2]);
#endif
		}

	      // Last message?
	      if ((isdigit ((int) line[0])) && (isdigit ((int) line[1])) && (isdigit ((int) line[2])) && (line[3] == ' '))
		{
		  parse = PARSE_SERVER_BEGIN;
		  release = 1;
		  ftp_answerstate (self);
		  parse = PARSE_SERVER_BEGIN;
		}

	      // Cuting the oversize...
	      if (length >= self->max_line_length)
		{
#ifdef ENABLE_DEBUG
		  z_debug (7, "%s: Csonkolok!", self->super.session_id);
#endif
		  length = self->max_line_length;
		}

	      // We need drop the lines :
	      //   - PASV, linear, NAT
	      //   - PASV, PORT->PASS, NAT
	      //   - PORT, PORT->PASS
	      //   - PASV, PASV->PORT
	      if (((((!strncmp (ansh, "04227PASV", strlen (ansh))) && (self->client_portpasv == 0) && (!self->last_port)) ||
		    ((!strncmp (ansh, "04227PASV", strlen (ansh))) && (self->client_portpasv == 2) && (!self->last_port))) && (self->NAT) && (self->internal_server))
		  || ((!strncmp (ansh, "04227PASV", strlen (ansh))) && (self->client_portpasv == 2) && (self->last_port))
		  || ((!strncmp (ansh, "04200PORT", strlen (ansh))) && (self->client_portpasv == 1) && (!self->last_port)))
		{
		  // Drop the remained lines
		  if (self->log_answer)
		    {
		      while (!ftp_log (self, LOG_ANSW + LOG_DROP, line, length));
		    }
		}
	      else
		{

		  // verbose mode -> log all
		  if (self->verbose_messages)
		    {
		      if (!nont)
			if (ftp_stream_write (self, 'C', line, length))
			  {
#ifdef ENABLE_DEBUG
			    z_debug (7, "%s: FTP: send failed %d %d", self->super.session_id, bytes_written, length + 2);
#endif
			    parse = PARSE_RETURN;
			    release = 1;
			    self->state = FTP_MAIN_RET;
			    break;
			  }

		      if (self->log_answer)
			{
			  while (!ftp_log (self, LOG_ANSW + LOG_PASS, line, length));
			}
		    }

		  // sort answers and the last one > interpret
		  if ((!self->verbose_messages) && (release))
		    {
		      if (!nont)
			if (ftp_stream_write (self, 'C', line, 4))
			  {
#ifdef ENABLE_DEBUG
			    z_debug (7, "%s: FTP: send failed %d 6", self->super.session_id, bytes_written);
#endif
			    parse = PARSE_RETURN;
			    release = 1;
			    self->state = FTP_MAIN_RET;
			    break;
			  }
		      if (self->log_answer)
			{
			  while (!ftp_log (self, LOG_ANSW + LOG_PASS, line, length));
			}
		    }
		}
	    }
	  else
	    {
	      // bad sequence...
#ifdef ENABLE_DEBUG
	      z_debug (7, "%s: FTP: break in answer sequence!", self->super.session_id);
#endif
	      if ((self->log_answer) && (self->log_dropped))
		{
		  while (!ftp_log (self, LOG_ANSW + LOG_DROP, line, length));
		}
	      parse = PARSE_RETURN;
	      self->state = FTP_MAIN_RET;
	      break;

	    }
	  break;
	case PARSE_RETURN:
	  release = 1;
	  self->state = FTP_MAIN_RET;
	  break;
	default:
	  break;
	}
    }
  // Ftp state is QUIT, and we get 221 ->
  //   it's the end...
  if ((self->ftp_state == FTP_STATE_QUIT) && (!strncmp (self->answer_code, "221", 3)))
    {
      self->state = FTP_MAIN_RET;
      self->quit = 1;
    }

  // We finished the interpreting...
  // Answer begin with 1 - stay in answer interpreter
  if ((self->state == FTP_SERVER_TO_CLIENT) && (self->answer_code[0] != '1'))
    self->state = FTP_CLIENT_TO_SERVER;
#ifdef ENABLE_DEBUG
  z_debug (7, "%s:  new state %d", self->super.session_id, self->state);
#endif

  return 0;
}

// Function:    ftp_nontransparent
// In:          FtpProxy                proxy pointer
// Out:         int                       
// Desc:        ftp nontransparent auth
int
ftp_nontransparent (FtpProxy * self)
{
  int i;
  char *dst;
  gchar *line;
  guint length = 0;
  guint IACl = 0;
  guint bytes_written;
  int readback;
  GString * str;

  ftp_init_stream_client (self);
  if (self->quit)
    return 1;

  if (self->verbose_messages)
    {
      if (ftp_sendback (self, msgs[MSG_WELCOME]))
	return 1;
      if (self->log_answer)
	{
	  while (!ftp_log (self, LOG_ANSW + LOG_PASS + LOG_PROX, msgs[MSG_WELCOME], strlen (msgs[MSG_WELCOME])));
	}
    }
  else
    {
      ftp_stream_write (self, 'C', "220 ", 4);
      if (self->log_answer)
	{
	  while (!ftp_log (self, LOG_ANSW + LOG_PASS + LOG_PROX, "220 ", 4));
	}
    }
  readback = ftp_read_line_get (self, 'C', &line, &length, &IACl);

#ifdef ENABLE_DEBUG
  z_debug (7, "%s: get %d %.*s", self->super.session_id, length, (int) length, line);
#endif
  if (length == 0)
    {
      self->quit = 1;
      return 1;
    }

  dst = NULL;

  if (ST_ERROR (readback) != ST_OK)
    {
#ifdef ENABLE_DEBUG
      z_debug (7, "%s: FTP: bad delimer! \\r\\n", self->super.session_id);
#endif
      if (ftp_sendback (self, msgs[MSG_TERM_CRLF]))
	return 1;
      if (self->log_dropped)
	{
	  while (!ftp_log (self, LOG_COMM + LOG_DROP, line, length));
	}
      if (self->log_answer)
	{
	  while (!ftp_log (self, LOG_ANSW + LOG_PASS + LOG_PROX, msgs[MSG_TERM_CRLF], strlen (msgs[MSG_TERM_CRLF])));
	}
      self->quit = 1;
      return 1;
    }
  else
    {
      if (length >= self->max_line_length)
	{
	  length = self->max_line_length;
	}
      if (strncasecmp (line, "USER ", 5))
	{
#ifdef ENABLE_DEBUG
	  z_debug (7, "%s: FTP: bad command!", self->super.session_id);
#endif
	  if (ftp_sendback (self, msgs[MSG_USER_FIRST]))
	    return 1;
	  if (self->log_dropped)
	    {
	      while (!ftp_log (self, LOG_COMM + LOG_DROP, line, length));
	    }
	  if (self->log_answer)
	    {
	      while (!ftp_log (self, LOG_ANSW + LOG_PASS + LOG_PROX, msgs[MSG_USER_FIRST], strlen (msgs[MSG_USER_FIRST])));
	    }
	  self->quit = 1;
	  return 1;
	}
      else
	{
	  if ((strrchr (line, '@')) == NULL)
	    {
#ifdef ENABLE_DEBUG
	      z_debug (7, "%s: FTP: bad form!", self->super.session_id);
#endif
	      if (ftp_sendback (self, msgs[MSG_BAD_USERF]))
		return 1;
	      if (self->log_dropped)
		{
		  while (!ftp_log (self, LOG_COMM + LOG_DROP, line, length));
		}
	      if (self->log_answer)
		{
		  while (!ftp_log (self, LOG_ANSW + LOG_PASS + LOG_PROX, msgs[MSG_BAD_USERF], strlen (msgs[MSG_BAD_USERF])));
		}
	      self->quit = 1;
	      return 1;
	    }
	  else
	    {
	      str = g_string_sized_new(length+1);
	      if ((dst = g_new0 (char, length + 1)) == NULL)
		{
#ifdef ENABLE_DEBUG
		  z_debug (7, "%s: FTP: failed allocate %d ", self->super.session_id, length + 1);
#endif
		  self->quit = 1;
		  return 1;
		}
	      g_string_sprintf(str,"%.*s",(int)length,line);
	      strncpy (dst, strrchr (str->str, '@') + 1, length);
	      g_string_truncate(str,strrchr (str->str, '@')-str->str);
	      strncpy (self->username, str->str + 5, self->max_username_length);
#ifdef ENABLE_DEBUG
	      z_debug (7, "%s: FTP: -%s@%s-", self->super.session_id, self->username, dst);
#endif
	      ftp_stream_write (self, 'C', "331 ", 4);
	      if (self->log_answer)
		{
		  while (!ftp_log (self, LOG_ANSW + LOG_PASS, "331 ", 4));
		}

	      readback = ftp_read_line_get (self, 'C', &line, &length, &IACl);

#ifdef ENABLE_DEBUG
	      z_debug (7, "%s: get %d %.*s", self->super.session_id, length, (int) length, line);
#endif
	      if (length == 0)
		{
		  self->quit = 1;
		  return 1;
		}

	      if (ST_ERROR (readback) != ST_OK)
		{
#ifdef ENABLE_DEBUG
		  z_debug (7, "%s: FTP: bad delimer! \\r\\n", self->super.session_id);
#endif
		  if (ftp_sendback (self, msgs[MSG_TERM_CRLF]))
		    return 1;
		  if (self->log_dropped)
		    {
		      while (!ftp_log (self, LOG_COMM + LOG_DROP, line, length));
		    }
		  if (self->log_answer)
		    {
		      while (!ftp_log (self, LOG_ANSW + LOG_PASS + LOG_PROX, msgs[MSG_TERM_CRLF], strlen (msgs[MSG_TERM_CRLF])));
		    }
		}
	      else
		{
		  if (length >= self->max_line_length)
		    {
		      length = self->max_line_length;
		    }

	          g_string_sprintf(str,"%.*s",(int)length,line);

		  if (strncasecmp (str->str, "PASS ", 5))
		    {
#ifdef ENABLE_DEBUG
		      z_debug (7, "%s: FTP: bad command! \\r\\n", self->super.session_id);
#endif
		      if (ftp_sendback (self, msgs[MSG_USERPASS_FIRST]))
			return 1;
		      if (self->log_dropped)
			{
			  while (!ftp_log (self, LOG_COMM + LOG_DROP, line, length));
			}
		      if (self->log_answer)
			{
			  while (!ftp_log (self, LOG_ANSW + LOG_PASS + LOG_PROX, msgs[MSG_USERPASS_FIRST], strlen (msgs[MSG_USERPASS_FIRST])));
			}
		       
		      self->quit = 1;
		      return 1;
		    }
		  else
		    {
		      //*(strchr (line, '\r')) = '\0';
		      strncpy (self->password, str->str + 5, self->max_password_length);
#ifdef ENABLE_DEBUG
		      z_debug (7, "%s: FTP: -%s-", self->super.session_id, self->password);
#endif
		      // !!!! 
		      //   dst must be checked in event
		      // !!!!
		      z_policy_lock (self->super.thread);
		      i = z_policy_event (self->super.handler, "nontransparent", z_policy_var_build ("csss", 'C', dst, self->username, self->password));
		      z_policy_unlock (self->super.thread);
		      switch (i)
			{
			case Z_ACCEPT:
			  z_proxy_connect_server_event (&self->super, dst, 21);
			  ftp_init_stream_server (self);
			  if (self->quit)
			    return 1;

			  if (dst)
			    g_free (dst);
			  if ((dst = g_new0 (char, self->max_line_length)) == NULL)
			    {
#ifdef ENABLE_DEBUG
			      z_debug (7, "%s: FTP: failed allocate %d ", self->super.session_id, self->max_line_length);
#endif
			      self->quit = 1;
			      return 1;
			    }
			  self->ftp_state = FTP_STATE_CONNECT;
			  self->state = FTP_SERVER_TO_CLIENT;
			  if (ftp_thread_server (self, 1))
			    {
#ifdef ENABLE_DEBUG
			      z_debug (7, "%s: FTP: failed ftp_thread_server ", self->super.session_id);
#endif
			      self->quit = 1;
			      return 1;
			    }
			  snprintf (dst, self->max_line_length, "USER %s", self->username);
			  strcpy (self->command, "USER");
			  ftp_stream_write (self, 'S', dst, strlen (dst));
			  self->ftp_state = FTP_STATE_LOGIN_U;
			  self->state = FTP_SERVER_TO_CLIENT;
			  if (ftp_thread_server (self, 1))
			    {
#ifdef ENABLE_DEBUG
			      z_debug (7, "%s: FTP: failed ftp_thread_server", self->super.session_id);
#endif
			      self->quit = 1;
			      return 1;
			    }
			  snprintf (dst, self->max_line_length, "PASS %s", self->password);
			  strcpy (self->command, "PASS");
			  ftp_stream_write (self, 'S', dst, strlen (dst));
			  // clear the password!!!
			  bzero (self->password, strlen (self->password));
			  self->ftp_state = FTP_STATE_LOGIN_P;
			  self->state = FTP_SERVER_TO_CLIENT;
			  if (ftp_thread_server (self, 1))
			    {
#ifdef ENABLE_DEBUG
			      z_debug (7, "%s: FTP: failed ftp_thread_server", self->super.session_id);
#endif
			      self->quit = 1;
			      return 1;
			    }
			  snprintf (dst, self->max_line_length, "%.*s ", 3, self->answer_code);
			  ftp_stream_write (self, 'C', dst, strlen (dst));
			  break;
			case Z_UNSPEC:
			case Z_REJECT:
			  if ((bytes_written = ftp_command_drop (self, line, length, msgs[MSG_NO_USER_ACCESS])) != strlen (msgs[MSG_NO_USER_ACCESS]) + 2)
			    {
#ifdef ENABLE_DEBUG
			      z_debug (7, "%s: FTP: failed send %d %d", self->super.session_id, bytes_written, strlen (msgs[MSG_NO_USER_ACCESS]));
#endif
			      self->quit = 1;
			      return 1;
			    }
			  self->quit = 1;
			  break;
			case Z_ABORT:
			  self->quit = 1;
			  break;
			default:
			  break;
			}

		    }
		}

	    }
	}
    }

  if (dst)
    g_free (dst);

  g_string_free(str,TRUE);

  return 0;
}

// Function:    ftp_thread
// In:          gpointer                proxy pointer
// Out:         -                       
// Desc:        ftp proxy thread
void
ftp_thread (gpointer s)
{
  FtpProxy *self = (FtpProxy *) s;
  gchar *line;
  guint length = 0;
  guint IACl = 0;
  guint bytes_written;
  gchar buff[5];
  FTPCommandDescriptor *workp;
  ftp_hash_fv fv;
  int readback;


  //
  //  Proxy startup
  //  
  ZPROXY_SET_STATE (self, PS_CONFIG);
  z_proxy_config_event (&self->super);

  ZPROXY_SET_STATE (self, PS_STARTING_UP);
  z_proxy_startup_event (&self->super);

  ZPROXY_SET_STATE (self, PS_WORKING);

  //
  // State machine init: transparent or not ...
  //
  if (self->proxy_state == 0)
    {
      self->ftp_state = FTP_STATE_CONNECT;
      z_proxy_connect_server_event (&self->super, NULL, 0);
      ftp_init_stream_client (self);
      ftp_init_stream_server (self);
    }
  else
    {
      self->ftp_state = FTP_STATE_PRECONNECT;
    }

  // 
  // non-Transparent state init
  //    
  if (self->ftp_state == FTP_STATE_PRECONNECT)
    {
      if (ftp_nontransparent (self))
	self->quit = 1;
    }

  // 
  // thread main loop
  // 
  while (!self->quit)
    {
      switch (self->state)
	{
	case FTP_INITIAL:
	  // Just started
	  self->state = FTP_SERVER_TO_CLIENT;
#ifdef ENABLE_DEBUG
	  z_debug (7, "%s:  new state %d", self->super.session_id, self->state);
#endif

	case FTP_SERVER_TO_CLIENT:
	  if (ftp_thread_server (self, 0))
	    {
#ifdef ENABLE_DEBUG
	      z_debug (7, "%s: FTP: failed ftp_thread_server", self->super.session_id);
#endif
	      self->quit = 1;
	      self->state = FTP_MAIN_RET;
	    }
	  break;

	case FTP_CLIENT_TO_SERVER:
	  // FROM client TO server
	  readback = ftp_read_line_get (self, 'C', &line, &length, &IACl);

#ifdef ENABLE_DEBUG
	  z_debug (7, "%s: get %d %.*s", self->super.session_id, length, (int) length, line);
#endif
	  if (length == 0)
	    {
	      self->state = FTP_MAIN_RET;
	      break;
	    }
#ifdef ENABLE_DEBUG
	  z_debug (7, "%s: FTP: delimer %d - %d %d %d", self->super.session_id, length, line[length - 3], line[length - 2], line[length - 1]);
#endif
	  if (ST_ERROR (readback) != ST_OK)
	    {
#ifdef ENABLE_DEBUG
	      z_debug (7, "%s: FTP: bad delimer! \\r\\n", self->super.session_id);
#endif
	      if (ftp_sendback (self, msgs[MSG_TERM_CRLF]))
		{
		  break;
		}
	      if (self->log_dropped)
		{
		  while (!ftp_log (self, LOG_COMM + LOG_DROP, line, length));
		}
	      if (self->log_answer)
		{
		  while (!ftp_log (self, LOG_ANSW + LOG_PASS + LOG_PROX, msgs[MSG_TERM_CRLF], strlen (msgs[MSG_TERM_CRLF])));
		}

	      break;
	    }
	  else if (length >= self->max_line_length)
	    {
	      // cut the oversize...
#ifdef ENABLE_DEBUG
	      z_debug (7, "%s: FTP: long line", self->super.session_id);
#endif
	      if (ftp_sendback (self, msgs[MSG_LINE_LONG]))
		{
		  break;
		}
	      if (self->log_dropped)
		{
		  while (!ftp_log (self, LOG_COMM + LOG_DROP, line, length));
		}
	      if (self->log_answer)
		{
		  while (!ftp_log (self, LOG_ANSW + LOG_PASS + LOG_PROX, msgs[MSG_LINE_LONG], strlen (msgs[MSG_LINE_LONG])));
		}
	      break;
	    }
	  else if (length < 3)
	    {
	      // every command must be minimum 3 char, + \r\n
#ifdef ENABLE_DEBUG
	      z_debug (7, "%s: FTP: empty line", self->super.session_id);
#endif
	      if (ftp_sendback (self, msgs[MSG_SYNTAX]))
		{
		  break;
		}
	      if (self->log_dropped)
		{
		  while (!ftp_log (self, LOG_COMM + LOG_DROP, line, length));
		}
	      if (self->log_answer)
		{
		  while (!ftp_log (self, LOG_ANSW + LOG_PASS + LOG_PROX, msgs[MSG_SYNTAX], strlen (msgs[MSG_SYNTAX])));
		}
	      break;
	    }

#ifdef ENABLE_DEBUG
	  z_debug (7, "%s: BEJUTOTT!", self->super.session_id);
#endif
	  // command looks good...
	  buff[3] = '\0';
	  buff[4] = '\0';
#ifdef ENABLE_DEBUG
	  z_debug (7, "%s:  %d %.*s", self->super.session_id, length, (int) length, line);
#endif
	  if (length < 5)
	    {
	      strncpy (buff, line, length);
	    }
	  else
	    {
	      strncpy (buff, line, 4);
	    }
	  buff[5] = '\0';
	  g_strchomp (buff);
	  g_strup(buff);
#ifdef ENABLE_DEBUG
	  z_debug (7, "%s: buff filled -%s- =%.*s=", self->super.session_id, buff, length, line);
#endif

	  // 
	  // FTP states
	  // 
	  switch (self->ftp_state)
	    {
	    case FTP_STATE_CONNECT:
	    case FTP_STATE_PRECONNECT:
#ifdef ENABLE_DEBUG
	      z_debug (7, "%s: FTP: ops, state-space error detected", self->super.session_id);
#endif
	      while (!ftp_log (self, LOG_PROX, "FTP: ops, state-space error detected", strlen ("FTP: ops, state-space error detected")));

	      self->ftp_state = FTP_STATE_RETURN;
	      self->state = FTP_MAIN_RET;
	      break;
	    case FTP_STATE_LOGIN:
	    case FTP_STATE_LOGIN_U:
	    case FTP_STATE_LOGIN_P:
	    case FTP_STATE_LOGIN_A:
	    case FTP_STATE_LOGINAUTH:
	      // HASH: good, but ... 
	      // we've got some state, and for reserving an int
	      // in the hash for the state information -> waste of mem. ???
	      if ((workp = (FTPCommandDescriptor *) ftp_command_hash_get (buff)) == NULL)
		{
#ifdef ENABLE_DEBUG
		  z_debug (7, "%s: NOT in HASH", self->super.session_id);
#endif
		  if ((bytes_written = ftp_command_drop (self, line, length, msgs[MSG_USERPASS_FIRST])) != strlen (msgs[MSG_USERPASS_FIRST]) + 2)
		    {
#ifdef ENABLE_DEBUG
		      z_debug (7, "%s: FTP: failed send %d %d", self->super.session_id, bytes_written, strlen (msgs[MSG_USERPASS_FIRST]));
#endif
		      self->quit = 1;
		      self->state = FTP_MAIN_RET;
		      break;
		    }
		}
	      else
		{
#ifdef ENABLE_DEBUG
		  z_debug (7, "%s: in HASH", self->super.session_id);
#endif
		  if ((fv = (ftp_hash_fv) (workp->fv)) == NULL)
		    {
#ifdef ENABLE_DEBUG
		      z_debug (7, "%s: FTP: bad function", self->super.session_id);
#endif
		      self->quit = 1;
		      self->state = FTP_MAIN_RET;
		      break;
		    }
		  if ((!strncmp (buff, "USER", 4)) ||
		      (!strncmp (buff, "PASS", 4)) || (!strncmp (buff, "ACCT", 4)) || (!strncmp (buff, "QUIT", 4)) || (!strncmp (buff, "HELP", 4)))
		    {
		      (*fv) (self, line, length, workp);
		    }
		  else
		    {
		      if ((bytes_written = ftp_command_drop (self, line, length, msgs[MSG_USERPASS_FIRST])) != strlen (msgs[MSG_USERPASS_FIRST]) + 2)
			{
#ifdef ENABLE_DEBUG
			  z_debug (7, "%s: FTP: failed send %d %d", self->super.session_id, bytes_written, strlen (msgs[MSG_USERPASS_FIRST]));
#endif
			  self->quit = 1;
			  self->state = FTP_MAIN_RET;
			  break;
			}
		    }
		}
	      break;
	    case FTP_STATE_CONVERSATION:
	      // In the connection ... We logged in ...
	      if ((workp = (FTPCommandDescriptor *) ftp_command_hash_get (buff)) == NULL)
		{
		  ftp_command_parse_UNKN (self, line, length, 0xFFFFFFFF);
		}
	      else
		{
		  if ((fv = (ftp_hash_fv) (workp->fv)) == NULL)
		    {
#ifdef ENABLE_DEBUG
		      z_debug (7, "%s: FTP: bad function", self->super.session_id);
#endif
		      self->quit = 1;
		      self->state = FTP_MAIN_RET;
		      break;
		    }
		  if ((*fv) (self, line, length, workp))
		    {
#ifdef ENABLE_DEBUG
		      z_debug (7, "%s: FTP: parser reported an error", self->super.session_id);
#endif
		      self->quit = 1;
		      self->state = FTP_MAIN_RET;
		      break;
		    }
		}
	      break;
	      // 
	      // The last state...
	      //
	    case FTP_STATE_QUIT:
	      self->state = FTP_SERVER_TO_CLIENT;
	      break;
	    case FTP_STATE_RETURN:
	      break;
	    default:
#ifdef ENABLE_DEBUG
	      z_debug (7, "%s: FTP: bad ftp_state", self->super.session_id);
#endif
	      self->quit = 1;
	      self->state = FTP_MAIN_RET;
	      break;
	    }

#ifdef ENABLE_DEBUG
	  z_debug (7, "%s:  new state %d", self->super.session_id, self->state);
#endif
	  break;
	case FTP_MAIN_RET:
	  self->quit = 1;
	  break;
	default:
#ifdef ENABLE_DEBUG
	  z_debug (7, "%s: FTP: bad state", self->super.session_id);
#endif
	  self->quit = 1;
	  self->state = FTP_MAIN_RET;
	  break;
	}
    }

  ZPROXY_SET_STATE (self, PS_SHUTTING_DOWN);
  z_proxy_shutdown_event (&self->super);

  z_proxy_destroy (&self->super);
  return;
}

// Function:    ftp_proxy_new
// In:          gchar                 session id
//              ZStream                 client struct.
//              ZPolicyObj              policy struct
// Out:         ZProxy*                 the proxy...                    
// Desc:        create a ftp proxy
ZProxy *
ftp_proxy_new (gchar * session_id, ZStream * client, ZPolicyObj * handler)
{
  FtpProxy *self = g_new0 (FtpProxy, 1);
  char buf[255];

  z_proxy_init (&self->super, session_id, client, handler);
  self->super.free = ftp_proxy_free;


  self->state = FTP_INITIAL;

  if (!ftp_command_hash)
    {
      ftp_create_hash (self);
#ifdef ENABLE_DEBUG
      z_debug (7, "%s: FTP: ftp_command_hash %d %d", self->super.session_id, g_hash_table_size (ftp_command_hash), sizeof (struct _FTPCommandDescriptor));
#endif
    }

  if (!ftp_answer_hash)
    {
      ftp_create_answer_hash (self);
#ifdef ENABLE_DEBUG
      z_debug (7, "%s: FTP: ftp_answer_hash %d %d", self->super.session_id, g_hash_table_size (ftp_answer_hash), sizeof (struct _FTPAnswerDescriptor));
#endif
    }

  // init variables
  if (!ftp_create_conf (self))
    return NULL;

  ftp_proxy_regvars (self);

  snprintf (buf, sizeof (buf), "ftp thread (%s)", client->name);
  z_thread_new (buf, ftp_thread, self);
  return &self->super;
}

// Function:    ftp_answer_descriptor_free
// In:          gpointer                key
//              gpointer                value
//              gpointer                user_data
// Out:         -
// Desc:        free the allocated memory
void
ftp_answer_descriptor_free (gpointer key, gpointer value, gpointer user_data)
{
  g_free (((FTPAnswerDescriptor *) value)->info);
  g_free ((FTPAnswerDescriptor *) value);
}

// Function:    ftp_proxy_free
// In:          ZProxy                  proxy struct.
// Out:         -
// Desc:        free the allocated memory, called from the core
void
ftp_proxy_free (ZProxy * s)
{
  FtpProxy *self = (FtpProxy *) s;
  int i;

#ifdef ENABLE_DEBUG
  z_debug (7, "%s: FTP: destroy", self->super.session_id);
#endif

  for (i = EP_CLIENT; i < EP_MAX; i++)
    {
      z_read_line_destroy (self->readline[i]);
      self->readline[i] = NULL;
    }

  // No dynamic element
  g_hash_table_destroy (self->local_command);
  // Dynamic elemnt...
  g_hash_table_foreach_remove (self->local_answer, (GHRFunc) ftp_answer_descriptor_free, NULL);
  g_hash_table_destroy (self->local_answer);

  g_free (self->username);
  g_free (self->password);
  g_free (self->parameter);
  g_free (self->path);

  g_string_free (self->client_data_address, 1);
  g_string_free (self->server_data_address, 1);
  g_string_free (self->fw_external_data_address, 1);
  g_string_free (self->fw_internal_data_address, 1);
  g_string_free (self->tmp_data_address, 1);

#ifdef ENABLE_DEBUG
  z_debug (7, "%s: FTP: by", self->super.session_id);
#endif
}

gint
zorp_module_init (void)
{
  z_registry_add ("ftp", ZR_PROXY, ftp_proxy_new);
  return TRUE;
}
