############################################################################
##
## Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
##
## $Id: Listener.py,v 1.15 2000/08/01 10:10:37 bazsi Exp $
##
############################################################################

from Zorp import *
import Zorp, Session, Service

class Listen:
	"""
	   Az osztly feladata, hogy a megadott porton figyeljen, es miutn
	   kapcsolat rkezett egy klienstl elindtsa a megadott szolgltatst.
	   
	   Mezlers:
	   	listen	A Zorp-ban lv alacsonyszint listen objektum 
						referencija.
	   	service	Az indtand szolgltats referencija (Service.Service).
	"""
	listen = None  
	service = None

	def __init__(self, bindto, service):
		"""
		   Ltrehoz egy Listen pldnyt.
		   
		   Paramterek:
		   	bindto		a bindoland cm
		   	service		az indtand szolgltats (Service.Service)

			Lers:
				Ltrehozza a Listen pldnyt s belltja a 
		"""
		self.service = service
	        self.listen = Zorp.Listen(bindto, self.accepted, 0)
	        self.local = self.listen.local
	        self.listen.start()

	def accepted(self, client, fd):
		"""
		   Callback, amit a core hv meg, amikor egy kapcsolat rkezett.
		   Ltrehoz, s elindt egy session-t.
		   
		   Paramterek:
		   	client		A kliens cme (SockAddr)
		   	fd		A kliens fel vezet kapcsolat fd-je.
		   	
		   Lers:
		   	Belerja a logba, hogy elfogadta a kapcsolatot, majd elinditja
				a megadott szolgltatst, figyelve arra, hogy a kapcsolat
				ip cim, s MAC/DAC policy alapjn engedlyezve van-e.

		   Visszatrsi rtk:
		   	int		Jelzi, hogy a kapcsolat tovbbmehet-e.		   	
		"""
		session = None
		try:
			session = Session.MasterSession()
			
			session.setService(self.service)
			session.setClient(fd, client)

			#service = self.getService(session)
			
			debug(10, "accepted a connection from python, client=%s" % client.format())
			session.service.startInstance(session)
			return Zorp.Z_ACCEPT
		except ZoneException, s:
			message(0, "Zone not found: %s" % (s))
		except DACException:
			message(0, "DAC policy violation")
		except MACException:
			message(0, "MAC policy violation")
		except AuthException:
			message(0, "Authentication failure")
		if session != None: 
			session.destroy()

		return Z_REJECT

	def getService(self, session):
		return self.service

	def destroy(self):
		"""
		   Lelltja az adott port figyelst.

			Paramterek:
				-

			Lers:
				-

		   Visszatrsi rtk:
		   	-
		"""
		self.listen.destroy()

#
#
#class MultiListen(Listen):
#	"""
#		A kliens znjtl fgg szolgltats indtsa.
#		
#		Mezlers:
#			services	A zna nevvel indexelt hash, mely 
#					megadja az alkalmazand szolgltatst.
#	"""
#	services = {}
#	
#	def __init__(self, bindto, services, default_service):
#		Listen.__init__(self, bindto, default_service)
#		self.services = services
#	
#	def getService(self, session):
#		try:
#			return self.services[session.client_zone];
#		except KeyError:
#			return self.service
#
