############################################################################
##
## Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
##
## $Id: policy.py,v 1.29 2000/08/08 16:48:57 bazsi Exp $
##
############################################################################

from Zorp import Zorp, SockAddr, Listener, Zone, Service, Chainer, Stream
from Zorp import Plug, POP3, AnyPy, Ftp, Session, Sink, Receiver, Http, Auth
from Zorp.Zorp import *
from Zorp.LabelSet import LabelSet

Zorp.firewall_name = 'bzorp@balabit'

Zorp.zones = \
[
	Zone.InetZone("balabit", "192.168.131.0", "255.255.255.0", None,
		outbound_services=["*"],
		inbound_services=["*"]),
        Zone.InetZone("local", "127.0.0.0", "255.0.0.0", None,
        	inbound_services=["*"],
        	outbound_services=["*"]),
        Zone.InetZone("internet", "0.0.0.0", "0.0.0.0", None,
        	inbound_services=["*"],
        	outbound_services=["*"])
]

class MySink(Sink.SinkProxy):

	def __init__(self, session):
		Zorp.debug(0, "MySink __init__()")
		Sink.SinkProxy.__init__(self, session)

	def config(self):
		Zorp.debug(0, "MySink config()")
		
	def startUp(self):
		Zorp.debug(0, "MySink startUp()")
		
	def packet(self, addr1, addr2, length):
		Zorp.debug(0, "MySink packet, length=%d, from=%s, to=%s" %
			      (length, addr1.format(), addr2.format())) 
		addr2.ip = "192.168.131.37"
		addr2.port = 1235
		return Zorp.Z_ACCEPT
		
class MyPlug(Plug.PlugProxy):

        class InnerPlug(Plug.PlugProxy):
                def startUp(self):
                        Zorp.debug(0, "innerplug startup")

        def config(self):
                self.copy_to_server = 0
                #self.stack_proxy = self.InnerPlug

	def startUp(self):
		Zorp.message(0, "Authenticated user: %s" % (self.session.auth_user))
		pass
				
	def shutDown(self):
		pass		
		
	def packetStats(self, client_bytes, client_pkt, server_bytes, server_pkt):
		self.aaa = 5
		Zorp.debug(0, "server->client: packet=%d, bytes=%d, bandwidth=%f" % (client_pkt, client_bytes, self.bandwidth_to_client))
		Zorp.debug(0, "client->server: packet=%d, bytes=%d, bandwidth=%f" % (server_pkt, server_bytes, self.bandwidth_to_server))
		return 1
		

class MyHttp(Http.HttpProxy):

        def config(self):
                Zorp.debug(0, "http.config()")
                #self.request_commands["POST"] = ()

        def startUp(self):
                Zorp.debug(0, "http.startup()")

        def shutDown(self):
                Zorp.debug(0, "http.shutdown()")

        def sendLog(self, type, msg):
                Zorp.debug(0, msg)

class MyTransHttp(Http.HttpProxy):

        def config(self):
                #self.server_address_transparent = "127.0.0.1"
                self.transparent_mode = 1
                Zorp.debug(0, "http.config()")

        def startUp(self):
                Zorp.debug(0, "http.startup()")

        def shutDown(self):
                Zorp.debug(0, "http.shutdown()")

        def sendLog(self, type, msg):
                Zorp.debug(0, msg)

class MyTransProxyHttp(Http.HttpProxy):

        def config(self):
                self.transparent_mode = 1
                Zorp.debug(0, "http.config()")

        def startUp(self):
                Zorp.debug(0, "http.startup()")

        def shutDown(self):
                Zorp.debug(0, "http.shutdown()")

        def sendLog(self, type, msg):
                Zorp.debug(0, msg)


class MyFtp(Ftp.FtpProxy):
	def startUp(self):
		#Zorp.debug(0, "MyFtp startUp()")
		self.command_accept = 0x00000000
                self.FtpCommandDISABLE("LIST")
                self.FtpCommandDISABLE("PORT")
                self.FtpAnswerENABLE("500", "CLNT")
                self.NAT = 1
		#self.fw_external_data_address="152.66.225.231"
		self.fw_internal_data_address="152.66.224.40"
                #self.proxy_state = 1

	def config(self):
		self.command_accept = 0x00000000
		
        def portCommand(self, dir, hashv, comm, param, ip, port):
                self.FtpCommandENABLE("LIST")
                return Z_ACCEPT

		
class MyFtpNonT(Ftp.FtpProxy):
	def startUp(self):
		#Zorp.debug(0, "MyFtp startUp()")
		self.command_accept = 0x00000000
                self.FtpCommandDISABLE("LIST")
                self.FtpCommandDISABLE("PORT")
                self.FtpAnswerENABLE( "500", "CLNT" )
                #self.NAT = 1
		#self.fw_external_data_address="152.66.225.231"
		#self.fw_internal_data_address="152.66.225.231"
                self.proxy_state = 1

	def config(self):
		self.command_accept = 0x00000000
		
		
def init(name):
	Zorp.debug(0, "Policy init, name=%s" % name)

	authconn = ZAS.AuthConnection(SockAddr.SockAddrInet('192.168.131.1', 1105), 'o=BalaBit IT Ltd.,c=HU')
	#auth = Satyr.SatyrAuthentication(authconn)
	
	plug_service = Service.Service("testplug", Chainer.DirectedChainer(None, SockAddr.SockAddrInet('127.0.0.1', 1998)), MyPlug, auth)
	plug_transparent_service = Service.Service("transplug", Chainer.TransparentChainer(None, 1), MyPlug)
	plug_failover = \
		Service.Service("failoverplug", 
			Chainer.FailoverChainer(None, 
				[SockAddr.SockAddrInet('127.0.0.1', 1234),
				 SockAddr.SockAddrInet('127.0.0.1', 1235)]),
			MyPlug)
	sink_service = Service.Service("sink", None, MySink)
	Listener.Listen(SockAddr.SockAddrInet("0.0.0.0", 1999), plug_failover)
	Listener.Listen(SockAddr.SockAddrInet("0.0.0.0", 2000), plug_transparent_service)
 	#ftp_service = Service.Service("ftp", Chainer.DirectedChainer(None, SockAddr.SockAddrInet('127.0.0.1', 21)), MyFtp)
 	ftp_service = Service.Service("ftp", Chainer.DirectedChainer(None, SockAddr.SockAddrInet('127.0.0.1', 21)), MyFtp)
 	ftp_trans = Service.Service("ftp", Chainer.TransparentChainer(None, 0), MyFtp)
 	ftp_nontr = Service.Service("ftp", Chainer.InbandChainer(None), MyFtpNonT)
 	Listener.Listen(SockAddr.SockAddrInet("0.0.0.0", 2010), ftp_service)
 	Listener.Listen(SockAddr.SockAddrInet("0.0.0.0", 2011), ftp_trans)
 	Listener.Listen(SockAddr.SockAddrInet("0.0.0.0", 2012), ftp_nontr)
	#Receiver.Receive(SockAddr.SockAddrInet("0.0.0.0", 1234), sink_service)

        # kliens szamara nem transzparens
        http_service = Service.Service("http", Chainer.InbandChainer(None), MyHttp)
        # kliens szerver kozott transzparens
        http_trans_service = Service.Service("http", Chainer.TransparentChainer(None, 1), MyTransHttp)
        # kliens proxy kozott, transzparensen
        http_trans_proxy_service = Service.Service("http", Chainer.TransparentChainer(None, 1), MyTransProxyHttp)

        Listener.Listen(SockAddr.SockAddrInet("0.0.0.0", 8080), http_service)
        Listener.Listen(SockAddr.SockAddrInet("0.0.0.0", 8081), http_trans_service)
        Listener.Listen(SockAddr.SockAddrInet("0.0.0.0", 8082), http_trans_proxy_service)


Zorp.debug(0, "Policy bootstrap done...");
